package com.zanxiang.game.back.serve.utils;

import com.github.sd4324530.jtuple.Tuple2;
import com.github.sd4324530.jtuple.Tuples;
import com.zanxiang.game.back.serve.pojo.entity.GameBackPolicy;
import com.zanxiang.game.back.serve.pojo.enums.BackUnitEnum;
import com.zanxiang.module.util.DateUtil;
import com.zanxiang.module.util.JsonUtil;
import com.zanxiang.module.util.NumberUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.RandomUtils;

import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.Objects;
import java.util.function.Function;

@Slf4j
public class BackPolicyUtil {
    private static final long[] RECHARGE_LEVEL = {12 * 100L, 30 * 100L, 50 * 100L, 98 * 100L, 198 * 100L, 328 * 100L, 649 * 100L, 998 * 100L, 2998 * 100L};

    static {
        Arrays.sort(RECHARGE_LEVEL);
    }

    /**
     * 订单回传
     *
     * @param gameBackPolicy   回传策略
     * @param rechargeMoney    充值金额
     * @param lastBackFunction 在指定最近订单数量内是否有订单回传过
     * @return <是否回传, 回传金额>
     */
    public static boolean backOrder(String orderId, GameBackPolicy gameBackPolicy, long rechargeMoney, boolean isFirstOrder, LocalDateTime payTime, LocalDateTime regTime, LastBackFunction lastBackFunction) {
        if (null == gameBackPolicy) {
            log.error("订单 id[{}]通过回传策略[null]判断后，是否回传：{}.", orderId, true);
            return true;
        }
        BackUnitEnum backUnit = gameBackPolicy.getBackUnit() == null ? BackUnitEnum.UNIT_ONCE : BackUnitEnum.getByValue(gameBackPolicy.getBackUnit());

        Boolean firstPolicy = null, firstDay = null;
        Long markUpTime = null;
        Integer backType;
        String backRate;
        if (gameBackPolicy.getRegPayIntervalTime() != null
                && ((DateUtil.localDateTimeToSecond(payTime) - DateUtil.localDateTimeToSecond(regTime)) / 60) > gameBackPolicy.getRegPayIntervalTime()) {
            if (gameBackPolicy.getMarkUpOrder() != null && gameBackPolicy.getMarkUpOrder()) {
                markUpTime = gameBackPolicy.getRegPayIntervalTime();
                if (rechargeMoney <= gameBackPolicy.getMarkUpOrderMinMoney()) {
                    // 小额
                    backType = gameBackPolicy.getMarkUpOrderMinMoneyType();
                    backRate = gameBackPolicy.getMarkUpOrderMinMoneyRate();
                } else if (rechargeMoney >= gameBackPolicy.getMarkUpOrderMaxMoney()) {
                    // 大额
                    backType = gameBackPolicy.getMarkUpOrderMaxMoneyType();
                    backRate = gameBackPolicy.getMarkUpOrderMaxMoneyRate();
                } else {
                    backType = gameBackPolicy.getMarkUpOrderOtherMoneyType();
                    backRate = gameBackPolicy.getMarkUpOrderOtherMoneyRate();
                }
                log.error("订单 id[{}]通过回传策略[{}]判断后，超过注册充值间隔时间，走补单逻辑 {}.", orderId, gameBackPolicy.getId(), gameBackPolicy.getRegPayIntervalTime());
            } else {
                log.error("订单 id[{}]通过回传策略[{}]判断后，超过注册充值间隔时间，不回穿 {}.", orderId, gameBackPolicy.getId(), gameBackPolicy.getRegPayIntervalTime());
                return false;
            }
        } else {
            if (backUnit == BackUnitEnum.UNIT_ONCE) {
                if (isFirstOrder) {
                    firstPolicy = true;
                    if (rechargeMoney <= gameBackPolicy.getFirstMinMoney()) {
                        // 小额
                        backType = gameBackPolicy.getFirstMinMoneyType();
                        backRate = gameBackPolicy.getFirstMinMoneyRate();
                    } else if (rechargeMoney >= gameBackPolicy.getFirstMaxMoney()) {
                        // 大额
                        backType = gameBackPolicy.getFirstMaxMoneyType();
                        backRate = gameBackPolicy.getFirstMaxMoneyRate();
                    } else {
                        backType = gameBackPolicy.getFirstOtherMoneyType();
                        backRate = gameBackPolicy.getFirstOtherMoneyRate();
                    }
                    log.error("订单 id[{}]通过回传策略[{}]判断后，正常回传. 首单：{}, backType: {}, backRate: {}", orderId, gameBackPolicy.getId(), true, backType, backRate);
                } else {
                    firstPolicy = false;
                    if (rechargeMoney <= gameBackPolicy.getRechargeMinMoney()) {
                        // 小额
                        backType = gameBackPolicy.getRechargeMinMoneyType();
                        backRate = gameBackPolicy.getRechargeMinMoneyRate();
                    } else if (rechargeMoney >= gameBackPolicy.getRechargeMaxMoney()) {
                        // 大额
                        backType = gameBackPolicy.getRechargeMaxMoneyType();
                        backRate = gameBackPolicy.getRechargeMaxMoneyRate();
                    } else {
                        backType = gameBackPolicy.getRechargeOtherMoneyType();
                        backRate = gameBackPolicy.getRechargeOtherMoneyRate();
                    }
                    log.error("订单 id[{}]通过回传策略[{}]判断后，正常回传. 首单：{}, backType: {}, backRate: {}", orderId, gameBackPolicy.getId(), false, backType, backRate);
                }
            } else if (backUnit == BackUnitEnum.UNIT_DAY) {
                long intervalDays = DateUtil.intervalOfDays(regTime.toLocalDate(), payTime.toLocalDate());
                if (intervalDays == 0) {
                    firstDay = true;
                    if (rechargeMoney <= gameBackPolicy.getFirstMinMoney()) {
                        // 小额
                        backType = gameBackPolicy.getFirstMinMoneyType();
                        backRate = gameBackPolicy.getFirstMinMoneyRate();
                    } else if (rechargeMoney >= gameBackPolicy.getFirstMaxMoney()) {
                        // 大额
                        backType = gameBackPolicy.getFirstMaxMoneyType();
                        backRate = gameBackPolicy.getFirstMaxMoneyRate();
                    } else {
                        backType = gameBackPolicy.getFirstOtherMoneyType();
                        backRate = gameBackPolicy.getFirstOtherMoneyRate();
                    }
                    log.error("订单 id[{}]通过回传策略[{}]判断后，正常回传. 首日：{}, backType: {}, backRate: {}", orderId, gameBackPolicy.getId(), true, backType, backRate);
                } else {
                    firstDay = false;
                    if (rechargeMoney <= gameBackPolicy.getRechargeMinMoney()) {
                        // 小额
                        backType = gameBackPolicy.getRechargeMinMoneyType();
                        backRate = gameBackPolicy.getRechargeMinMoneyRate();
                    } else if (rechargeMoney >= gameBackPolicy.getRechargeMaxMoney()) {
                        // 大额
                        backType = gameBackPolicy.getRechargeMaxMoneyType();
                        backRate = gameBackPolicy.getRechargeMaxMoneyRate();
                    } else {
                        backType = gameBackPolicy.getRechargeOtherMoneyType();
                        backRate = gameBackPolicy.getRechargeOtherMoneyRate();
                    }
                    log.error("订单 id[{}]通过回传策略[{}]判断后，正常回传. 首日：{}, backType: {}, backRate: {}", orderId, gameBackPolicy.getId(), false, backType, backRate);
                }
            } else {
                throw new RuntimeException("未知的回传单位：" + backUnit);
            }
        }
        if (Objects.equals(backType, GameBackPolicy.POLICY_TYPE_FIXED_RATE)) {
            // 固定比例回传
            String[] temp = backRate.split(":");
            int backCount = Integer.parseInt(temp[0]);
            int ignoreCount = Integer.parseInt(temp[1]);
            if (backCount == 0) {
                log.error("订单 id[{}]通过回传策略[{}]判断后，是否回传：{}.", orderId, gameBackPolicy.getId(), false);
                return false;
            }
            if (ignoreCount == 0) {
                log.error("订单 id[{}]通过回传策略[{}]判断后，是否回传：{}.", orderId, gameBackPolicy.getId(), true);
                return true;
            }
            int isBackCount = lastBackFunction.getLastBackCount(backCount + ignoreCount - 1, firstPolicy, firstDay, markUpTime);
            log.error("订单 id[{}]通过回传策略[{}]判断后，是否回传：{}.", orderId, gameBackPolicy.getId(), (isBackCount < backCount));
            return isBackCount < backCount;
        } else if (Objects.equals(backType, GameBackPolicy.POLICY_TYPE_RANDOM_RATE)) {
            // 随机概率回传
            boolean isBack = RandomUtils.nextInt(0, 100) < NumberUtil.multiply100(new BigDecimal(backRate)).intValue();
            log.error("订单 id[{}]通过回传策略[{}]判断后，是否回传：{}.", orderId, gameBackPolicy.getId(), isBack);
            return isBack;
        } else {
            throw new RuntimeException("回传策略[" + gameBackPolicy.getId() + "]配置错误，未知的回传类型：" + backType);
        }
    }

    private static Tuple2<Integer, Integer> approximate(int molecular, int denominator) {
        if (molecular % denominator == 0) {
            int temp = molecular / denominator;
            return Tuples.tuple(temp, temp);
        }
        int model = 2;
        do {
            if (molecular % model == 0 && denominator % model == 0) {
                molecular = molecular / model;
                denominator = denominator / model;
                continue;
            }
            model++;
        } while (model <= denominator / 2);
        return Tuples.tuple(molecular, denominator);
    }

    /**
     * 降档
     */
    public static long lowRechargeLevel(long rechargeMoney, Integer levelDown) {
        if (levelDown == null || levelDown == 0) {
            return rechargeMoney;
        }
        int level = RECHARGE_LEVEL.length - 1;
        for (int i = 0; i < RECHARGE_LEVEL.length; i++) {
            if (rechargeMoney <= RECHARGE_LEVEL[i]) {
                level = i;
                break;
            }
        }
        level = level - levelDown;
        level = level < 0 ? 0 : Math.min(level, RECHARGE_LEVEL.length - 1);
        return RECHARGE_LEVEL[level];
    }

    public interface LastBackFunction {
        int getLastBackCount(int backUnit, Boolean isFirstOrder, Boolean firstDay, Long markUpTime);
    }
}
